//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2016
//#
//#
//# File Name:  adc_conversion_control.v
//# Author:     Earth People Technology
//# Date:       July 8, 2016
//# Revision:   A
//#
//# Development: EPT Digital Storage Oscilloscope Project 
//# Application: Altera Cyclone IV FPGA
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. 
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			7/8/16	1			Created
//#
//#									
//#
//#######################################################################

`resetall
`timescale 1ns/10ps

`include "../src/dso_define.v"



module adc_conversion_control(


	input  wire                 CLK,
   	input  wire                 RST_N,

	input   wire                ADC_CONVST_COMMAND_LATCH,
	input   wire [7:0]          ADC_CONVST_COMMAND,
	
	output  reg  [23:0]         ADC_CONVST_DELAY_REGISTER,
	
	inout   wire [7:0]          MEMORY_ADDRESS,
	inout   wire [7:0]          MEMORY_DATA,
	output  wire                MEMORY_WRITE_LATCH,
	output  wire                MEMORY_READ_LATCH,
	output  wire                MEMORY_OUTPUT_ENABLE,
	
	output  reg                 REQUEST_EN,
	input   wire                GRANT_RDY
	
	

 		);



//==================================================================
// Parameter Declarations
//==================================================================
	parameter	IDLE                       = 0,
                START                      = 1,
                LATCH_COMMAND              = 2,
                CHECK_BYTE_COUNT           = 3,
                MEMORY_RD_ASSERT_OE        = 4,
                MEMORY_RD_ADDRESS          = 5,
                MEMORY_RD_DATA             = 6,
                MEMORY_RD_LATCH            = 7,
				MEMORY_RD_COMPLETE         = 8,
                ADC_CONVST_DELAY_COMPLETE  = 9;

				

//==================================================================
// Register Declarations
//==================================================================

	// State Machine Registers
	reg	[9:0]                  state;
	reg	[9:0]                  next;

	//Memory Latch registers
	reg                         memory_rd_address_flag;
	
	//Memory Write registers
	reg                         memory_rd_latch_flag;
	reg                         memory_rd_data_flag;
	reg                         memory_rd_latch_reg;
	reg                         memory_output_enable_flag;	
	
	//ADC CONVST Delay Command Storage from EPT-4CE6-Top
	reg [7:0]                   adc_convst_delay_start_address_in_memory;
	reg [7:0]                   adc_convst_delay_command_storage;	
	reg                         adc_convst_delay_reg;
	reg [7:0]                   number_of_adc_convst_bytes_to_read;
	
	//Local Memory Grant Request Registers
	reg [1:0]                   request_en_reg;	
	
	//Memory Read Cycle Wait State
	reg [7:0]                   memory_rd_data_counter;
	

//==================================================================
// Simulation Directives
//==================================================================

`ifdef SIM
	reg [8*30:1] state_name;
`endif

//=================================================================================
// Assignments	
//=================================================================================

	assign          MEMORY_ADDRESS = (memory_rd_address_flag) ? adc_convst_delay_start_address_in_memory: 8'hzz;
	//assign          //MEMORY_DATA = memory_wr_data_flag ? serial_receive_out_byte : 8'hzz;
	assign          MEMORY_WRITE_LATCH = 1'b0;
	assign          MEMORY_READ_LATCH = memory_rd_latch_flag ? 1'b1 : 1'b0;
	assign          MEMORY_OUTPUT_ENABLE = memory_output_enable_flag ? 1'b1 : 1'b0;


//=================================================================================
//  Update the ADC CONVST Delay Register Cycle
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
 `ifdef SIM
            ADC_CONVST_DELAY_REGISTER <= 24'h500;
`else
	        ADC_CONVST_DELAY_REGISTER <= 19'h000ce4; //0.07ms
	        //ADC_CONVST_DELAY_REGISTER <= 16'h03de; //0.034ms
	        //ADC_CONVST_DELAY_REGISTER <= 24'ha1220; //100.0ms
`endif
		   adc_convst_delay_reg <= 1'b0;
		end
  		else
    	begin
		   if(state[MEMORY_RD_LATCH])// & !adc_convst_delay_reg)
		   begin
		      case(number_of_adc_convst_bytes_to_read)
			  8'h3:
		         ADC_CONVST_DELAY_REGISTER[7:0] <= MEMORY_DATA;
			  8'h2:
		         ADC_CONVST_DELAY_REGISTER[15:8] <= MEMORY_DATA;
			  8'h1:
		         ADC_CONVST_DELAY_REGISTER[23:16] <= MEMORY_DATA;
			  endcase
		      adc_convst_delay_reg <= 1'b1;
		   end
		   else if((state[CHECK_BYTE_COUNT] | ADC_CONVST_DELAY_COMPLETE))// & adc_convst_delay_reg)
		     adc_convst_delay_reg <= 1'b0;
    	end
  	end

		
 	
//=================================================================================
//  ADC CONVST Delay Register Command Cycle: Read in the command from the EPT-4CE6-Top
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   adc_convst_delay_command_storage <= 0;
		end
  		else
    	begin
		   if(ADC_CONVST_COMMAND_LATCH)
		      adc_convst_delay_command_storage <= ADC_CONVST_COMMAND;
    	end
  	end

//=================================================================================
//  ADC CONVST Delay Register Command Cycle: Latch the SPI Command to send over SPI
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
			number_of_adc_convst_bytes_to_read <= 0;
			adc_convst_delay_start_address_in_memory <= 0;
		    memory_rd_latch_reg <= 1'b0;
		end
  		else
    	begin
		   if(state[LATCH_COMMAND])
		   begin
		      case(adc_convst_delay_command_storage[7:4])
			  4'h1:
			  begin
			     number_of_adc_convst_bytes_to_read <= 3'h3;
				 adc_convst_delay_start_address_in_memory <= `ADC_CONVST_DLY_1_MEMORY_LOCATION;//8'h16;
			  end
			  4'h2:
			  begin
			  end
			  endcase
		   end
           else if(state[MEMORY_RD_DATA])
		   begin
		      memory_rd_latch_reg <= 1'b0;
		   end
           else if(state[CHECK_BYTE_COUNT] & !memory_rd_latch_reg)
		   begin
		      memory_rd_latch_reg <= 1'b1;
		      adc_convst_delay_start_address_in_memory <= adc_convst_delay_start_address_in_memory + 1'd1;
		      if(number_of_adc_convst_bytes_to_read > 0)
		         number_of_adc_convst_bytes_to_read <= number_of_adc_convst_bytes_to_read - 1'd1;
		   end
           else if(state[IDLE] | (state[MEMORY_RD_ADDRESS] & !MEMORY_READ_LATCH & memory_rd_latch_reg))
		   begin
		      memory_rd_latch_reg <= 1'b0;
		   end
   	end
  	end

  
//=================================================================================
//  Memory Access Cycle: REQUEST_EN
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   REQUEST_EN <= 1'b0;
           request_en_reg <= 2'b00;
		end
  		else
    	begin
		   if(state[LATCH_COMMAND] | state[CHECK_BYTE_COUNT] & (request_en_reg == 2'b00))
		   begin
		      request_en_reg <= 2'b01;
		   end
		   else if(!GRANT_RDY & (request_en_reg == 2'b01))
		   begin
              request_en_reg <= 2'b10;
		      REQUEST_EN <= 1'b1;
		   end
		   else if(GRANT_RDY & (request_en_reg == 2'b10))
		   begin
              request_en_reg <= 2'b11;
		      REQUEST_EN <= 1'b1;
		   end
		   else if(state[ADC_CONVST_DELAY_COMPLETE] & (request_en_reg == 2'b11))
		   begin
              request_en_reg <= 2'b00;
		      REQUEST_EN <= 1'b0;
		   end
    	end
  	end

		
  
//=================================================================================
//  Read Cycle: Memory Output Enable
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   memory_output_enable_flag <= 1'b0;
		end
  		else
		   if(state[MEMORY_RD_ASSERT_OE])
		      memory_output_enable_flag <= 1'b1;
		   else if(state[MEMORY_RD_COMPLETE])
		      memory_output_enable_flag <= 1'b0;
   	end

//=================================================================================
//	Read Cycle: Memory Read Address
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_rd_data_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_RD_DATA])
		      memory_rd_data_flag <= 1'b1;
		   else if(state[MEMORY_RD_COMPLETE])
		      memory_rd_data_flag <= 1'b0;
		end
	end

//=================================================================================
//	Read Cycle: Memory Read Data
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_rd_address_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_RD_ASSERT_OE])
		   begin
		      memory_rd_address_flag <= 1'b1;
		   end
		   else if(state[MEMORY_RD_COMPLETE])
		      memory_rd_address_flag <= 1'b0;
		end
	end
	
//=================================================================================
//	Read Cycle: Memory Read Latch
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_rd_latch_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_RD_LATCH])
		   begin
		      memory_rd_latch_flag <= 1'b1;
		   end
		   else if(state[MEMORY_RD_COMPLETE])
		      memory_rd_latch_flag <= 1'b0;
		end
	end
			
//=================================================================================
//  Read Cycle: Wait States for MEMORY_RD_DATA
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   memory_rd_data_counter <= 0;
		end
  		else
		   if(state[MEMORY_RD_LATCH] | state[MEMORY_RD_COMPLETE])
		   begin
		      memory_rd_data_counter <= memory_rd_data_counter + 1'd1;
		   end
		   else if(state[CHECK_BYTE_COUNT])
		      memory_rd_data_counter <= 0;
   	end


//=================================================================================
//	Serial I2C Communications Finite State Machine
//              IDLE                       = 0,
//              START                      = 1,
//              LATCH_COMMAND              = 2,
//              CHECK_BYTE_COUNT           = 3,
//              MEMORY_RD_ASSERT_OE        = 4,
//              MEMORY_RD_ADDRESS          = 5,
//              MEMORY_RD_DATA             = 6,
//              MEMORY_RD_LATCH            = 7,
//              ADC_CONVST_DELAY_COMPLETE  = 8;
//
//=================================================================================

	always @(  posedge CLK or negedge RST_N  )
	begin
		if( !RST_N )
		begin
			state <= 10'h00;
			state [IDLE] <= 1'b1;
		end
		else
			state <= next;
	end


	always @( state or ADC_CONVST_COMMAND_LATCH or memory_output_enable_flag
	         or GRANT_RDY or number_of_adc_convst_bytes_to_read
			 or memory_rd_latch_flag or memory_rd_data_counter)
	begin
		next = 10'h00;
		if(state[IDLE])
		begin
			if(ADC_CONVST_COMMAND_LATCH)
				next[START] = 1'b1;
			else
				next[IDLE] = 1'b1;
		end

		if(state[START])
		begin
          next[LATCH_COMMAND] = 1'b1;
		end

		if(state[LATCH_COMMAND])
		begin
			next[MEMORY_RD_ASSERT_OE] = 1'b1;
		end

		if(state[CHECK_BYTE_COUNT])
		begin
			if(number_of_adc_convst_bytes_to_read > 0)
			begin
				next[MEMORY_RD_ASSERT_OE] = 1'b1;
			end
			else
				next[ADC_CONVST_DELAY_COMPLETE] = 1'b1;
		end

		if(state[MEMORY_RD_ASSERT_OE])
		begin
			if(GRANT_RDY)
				next[MEMORY_RD_ADDRESS] = 1'b1;
			else
				next[MEMORY_RD_ASSERT_OE] = 1'b1;
		end

		if(state[MEMORY_RD_ADDRESS])
		begin
           next[MEMORY_RD_DATA] = 1'b1;
		end

		if(state[MEMORY_RD_DATA])
		begin
           next[MEMORY_RD_LATCH] = 1'b1;
		end

		if(state[MEMORY_RD_LATCH])
		begin
		   if(memory_rd_data_counter < `MEMORY_RD_LATCH_WAIT_STATE)
		      next[MEMORY_RD_LATCH] = 1'b1;
		   else
		      next[MEMORY_RD_COMPLETE] = 1'b1;
		end

		if(state[MEMORY_RD_COMPLETE])
		begin
		   if(memory_rd_data_counter < `MEMORY_RD_COMPLETE_WAIT_STATE)
		      next[MEMORY_RD_COMPLETE] = 1'b1;
		   else
		      next[CHECK_BYTE_COUNT] = 1'b1;
		end

		if(state[ADC_CONVST_DELAY_COMPLETE])
		begin
                  next[IDLE] = 1'b1;
		end


`ifdef SIM
begin
	if ( state == (1 << IDLE))
		state_name = "IDLE";		
	if ( state == (1 << START))
		state_name = "START";		
	if ( state == (1 << LATCH_COMMAND))
		state_name = "LATCH_COMMAND";		
	if ( state == (1 << CHECK_BYTE_COUNT))
		state_name = "CHECK_BYTE_COUNT";		
	if ( state == (1 << MEMORY_RD_ASSERT_OE))
		state_name = "MEMORY_RD_ASSERT_OE";		
	if ( state == (1 << MEMORY_RD_ADDRESS))
		state_name = "MEMORY_RD_ADDRESS";
	if ( state == (1 << MEMORY_RD_DATA))
		state_name = "MEMORY_RD_DATA";
	if ( state == (1 << MEMORY_RD_LATCH))
		state_name = "MEMORY_RD_LATCH";
	if ( state == (1 << MEMORY_RD_COMPLETE))
		state_name = "MEMORY_RD_COMPLETE";
	if ( state == (1 << ADC_CONVST_DELAY_COMPLETE))
		state_name = "ADC_CONVST_DELAY_COMPLETE";
end
`endif		
	end

endmodule

