//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2016
//#
//#
//# File Name:  spi_communications.v
//# Author:     Earth People Technology
//# Date:       March 8, 2016
//# Revision:   A
//#
//# Development: EPT Digital Storage Oscilloscope Project 
//# Application: Altera Cyclone IV FPGA
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. 
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			3/8/16	1			Created
//#
//#									
//#
//#######################################################################

`resetall
`timescale 1ns/10ps

`include "../src/dso_define.v"



module spi_communications(


	input  wire                 CLK,
   	input  wire                 RST_N,

   	output wire                 SPI_DATA_OUT,
   	output wire                 SPI_CLK,
   	input  wire                 SPI_DATA_IN,
   	output wire                 SPI_CS_N,
	
	inout   wire [7:0]          MEMORY_ADDRESS,
	inout   wire [7:0]          MEMORY_DATA,
	output  wire                MEMORY_WRITE_LATCH,
	output  wire                MEMORY_READ_LATCH,
	output  wire                MEMORY_OUTPUT_ENABLE,
	
	input   wire [7:0]          SPI_COMMAND,
	input   wire                SPI_COMMAND_LATCH,
	
	output  reg                 SPI_TRANSMIT_COMPLETE,
	output  wire                SPI_BYTE_RECEIVED,

	output  reg                 REQUEST_EN,
	input   wire                GRANT_RDY
	
	

 		);



//==================================================================
// Parameter Declarations
//==================================================================
	parameter	IDLE                       = 0,
                START                      = 1,
                LATCH_COMMAND              = 2,
                CHECK_READ_WRITE_FLAG      = 3,
                MEMORY_RD_ASSERT_OE        = 4,
                MEMORY_RD_ADDRESS          = 5,
                MEMORY_RD_DATA             = 6,
                MEMORY_RD_LATCH            = 7,
                START_SPI_TRANSMIT         = 8,
				WAIT_FOR_SPI_COMPLETE      = 9,
                CHECK_SPI_COMPLETE         = 10,
				START_SPI_TRANSMIT_DELAY   = 11,
                START_SPI_RECEIVE          = 12,
                MEMORY_WRITE_CYCLE         = 13,
                MEMORY_WR_ADDRESS          = 14,
                MEMORY_WR_DATA             = 15,
                MEMORY_WR_LATCH            = 16;

				
    //PGA Commands
	//parameter   PGA_WRITE_COMMAND          = 8'h2a;
	//parameter   PGA_READ_COMMAND           = 8'h4a;

//==================================================================
// Register Declarations
//==================================================================

	// State Machine Registers
	reg	[16:0]                  state;
	reg	[16:0]                  next;

	//Receive Storage registers
	reg  [7:0]                  serial_receive_data_storage;	
    reg  [1:0]                  serial_receive_data_reg;
	wire [7:0]                  transfer_1_address;

	//Serial SPI Transmit Registers
	reg    [7:0]                spi_transfer_storage_register;
	wire                        serial_transmit_busy;
	wire                        master_spi_transmit_complete;
	reg                         master_spi_initiate_receive;
	wire                        master_spi_byte_received;
	wire   [7:0]                serial_receive_out_byte;
	reg                         master_spi_initiate_transmit;
	reg    [7:0]                spi_command_to_send;
	reg    [3:0]                number_of_spi_read_bytes;
	reg                         spi_read_write_flag;
	reg    [7:0]                spi_payload_bytes [3:0];
	reg                         spi_payload_byte_count_reg;
	wire                        spi_cs_from_spi_iface;

	//SPI Transmit Byte counters
	reg    [3:0]                number_of_spi_payload_bytes;
	reg    [3:0]                number_of_spi_bytes_to_send;
	reg    [3:0]                spi_payload_byte_count;	
	reg    [3:0]                spi_transmit_total_count;
	reg                         spi_transmit_total_count_reg;

	
	//SPI Receive Registers
	reg    [3:0]                spi_receive_total_count;
	reg                         start_spi_receive_reg;
	
	
	// Read Cycle Complete delay counter
	//reg [7:0]                   read_complete_counter;
	//reg                         read_complete_flag;
	
	//Memory Latch registers
	reg                         memory_wr_address_flag;
	reg                         memory_rd_address_flag;
	reg                         memory_wr_latch_reg;
	
	//Memory Write registers
	reg                         memory_wr_data_flag;
	reg                         memory_wr_latch_flag;
	reg                         memory_rd_latch_flag;
	reg                         memory_rd_data_flag;
	reg                         memory_output_enable_flag;	
	
	//Memory Interface registers
	reg [7:0]                   spi_payload_start_address_in_memory;
	
	//SPI Command Storage from UC Controller
	reg [7:0]                   spi_command_storage;	
	

//==================================================================
// Simulation Directives
//==================================================================

`ifdef SIM
	reg [8*30:1] state_name;
`endif

//=================================================================================
// Assignments	
//=================================================================================

	assign          MEMORY_ADDRESS = (memory_wr_address_flag | memory_rd_address_flag) ? spi_payload_start_address_in_memory: 8'hzz;
	assign          MEMORY_DATA = memory_wr_data_flag ? serial_receive_out_byte : 8'hzz;
	assign          MEMORY_WRITE_LATCH = memory_wr_latch_flag ? 1'b1 : 1'b0;
	assign          MEMORY_READ_LATCH = memory_rd_latch_flag ? 1'b1 : 1'b0;
	assign          MEMORY_OUTPUT_ENABLE = memory_output_enable_flag ? 1'b1 : 1'b0;
	
	assign          SPI_BYTE_RECEIVED = master_spi_byte_received;
	
	//Modify the SPI_CS signal from the spi_iface
	assign          SPI_CS_N = (state[START_SPI_TRANSMIT] | state[START_SPI_TRANSMIT_DELAY] | state[WAIT_FOR_SPI_COMPLETE] | state[MEMORY_WR_LATCH] | state[START_SPI_RECEIVE] | state[CHECK_SPI_COMPLETE] |state[MEMORY_WRITE_CYCLE]) ? 1'b0 : spi_cs_from_spi_iface;


//=================================================================================
//  SPI Command Cycle: Read in the command from the uc controller
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   spi_command_storage <= 0;
		end
  		else
    	begin
		   if(SPI_COMMAND_LATCH)
		      spi_command_storage <= SPI_COMMAND;
    	end
  	end

//=================================================================================
//  SPI Command Cycle: Latch the SPI Command to send over SPI
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
			number_of_spi_bytes_to_send <= 0;
			spi_command_to_send <= 0;
			spi_read_write_flag <= 1'b0;
			number_of_spi_payload_bytes <= 0;
			number_of_spi_read_bytes <= 3'h1;
			spi_payload_start_address_in_memory <= 0;
		    memory_wr_latch_reg <= 1'b0;
		end
  		else
    	begin
		   if(state[LATCH_COMMAND])
		   begin
		      case(spi_command_storage[7:4])
			  4'h8:
			  begin
			     number_of_spi_bytes_to_send <= 3'h2;
				 number_of_spi_payload_bytes <= 3'h1;
				 spi_command_to_send <= `PGA_WRITE_COMMAND;
				 spi_read_write_flag <= 1'b0;
				 case(spi_command_storage[3:0])
				 4'h1:
				     spi_payload_start_address_in_memory <= `PGA_CH1_MEMORY;//8'h10;
				 4'h2:
				     spi_payload_start_address_in_memory <= `PGA_CH2_MEMORY;//8'h12;
				 4'h3:
				     spi_payload_start_address_in_memory <= `PGA_CH3_MEMORY;//8'h14;
				 4'h4:
				     spi_payload_start_address_in_memory <= `PGA_CH4_MEMORY;//8'h16;
				 endcase
			  end
			  4'h9:
			  begin
				 number_of_spi_payload_bytes <= 3'h2;
				 number_of_spi_bytes_to_send <= 3'h2;
				 number_of_spi_read_bytes <= 3'h2;
				 spi_command_to_send <= `PGA_READ_COMMAND;
				 spi_read_write_flag <= 1'b1;
				 spi_payload_start_address_in_memory <= `PGA_READ_MEMORY_LOCATION;
			  end
			  endcase
		   end
           else if(state[CHECK_SPI_COMPLETE] & !MEMORY_WRITE_LATCH )
		   begin
		      memory_wr_latch_reg <= 1'b0;
		      //if(number_of_spi_bytes_to_send > 0)
		      //   number_of_spi_bytes_to_send <= number_of_spi_bytes_to_send - 1'd1;
			  if(number_of_spi_payload_bytes > 0)
                 number_of_spi_payload_bytes <= number_of_spi_payload_bytes - 1'd1;
		   end
           else if(state[CHECK_SPI_COMPLETE] & MEMORY_WRITE_LATCH & !memory_wr_latch_reg)
		   begin
		      memory_wr_latch_reg <= 1'b1;
		      spi_payload_start_address_in_memory <= spi_payload_start_address_in_memory + 1'd1;
		   end
           else if(state[IDLE] | (state[MEMORY_WRITE_CYCLE] & !MEMORY_WRITE_LATCH & memory_wr_latch_reg))
		   begin
		      memory_wr_latch_reg <= 1'b0;
		   end
   	end
  	end

  
//=================================================================================
//  Memory Access Cycle: REQUEST_EN
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   REQUEST_EN <= 1'b0;
		end
  		else
    	begin
		   if(state[CHECK_READ_WRITE_FLAG])
		      REQUEST_EN <= 1'b1;
		   else if(state[IDLE])//state[MEMORY_RD_ADDRESS] | state[MEMORY_WR_ADDRESS])
		     REQUEST_EN <= 1'b0;
    	end
  	end

		
  
//=================================================================================
//  Read Cycle: Memory Output Enable
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   memory_output_enable_flag <= 1'b0;
		end
  		else
		   if(state[MEMORY_RD_ASSERT_OE])
		      memory_output_enable_flag <= 1'b1;
		   else if(state[MEMORY_RD_LATCH])
		      memory_output_enable_flag <= 1'b0;
   	end

//=================================================================================
//	Read Cycle: Memory Read Address
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_rd_data_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_RD_DATA])
		      memory_rd_data_flag <= 1'b1;
		   else if(state[CHECK_SPI_COMPLETE])
		      memory_rd_data_flag <= 1'b0;
		end
	end

//=================================================================================
//	Read Cycle: Memory Read Data
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_rd_address_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_RD_ASSERT_OE])
		   begin
		      memory_rd_address_flag <= 1'b1;
		   end
		   else if(state[CHECK_SPI_COMPLETE])
		      memory_rd_address_flag <= 1'b0;
		end
	end
	
//=================================================================================
//	Read Cycle: Memory Read Latch
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_rd_latch_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_RD_LATCH])
		   begin
		      memory_rd_latch_flag <= 1'b1;
		   end
		   else if(state[START_SPI_TRANSMIT])
		      memory_rd_latch_flag <= 1'b0;
		end
	end
			
//=================================================================================
//	SPI Transmit Cycle: Start 
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   master_spi_initiate_transmit <= 1'b0;
		   spi_transmit_total_count <= 0;
		   spi_transmit_total_count_reg <= 1'b0;
		end
		else
		begin
		   if(state[START_SPI_TRANSMIT])
		   begin
		      master_spi_initiate_transmit <= 1'b1;
			  spi_transfer_storage_register <= spi_payload_bytes[spi_transmit_total_count];
		   end
		   else if(state[WAIT_FOR_SPI_COMPLETE] & !spi_transmit_total_count_reg)
		   begin
		      spi_transmit_total_count <= spi_transmit_total_count + 1'd1;
		      master_spi_initiate_transmit <= 1'b0;
			  spi_transmit_total_count_reg <= 1'b1;
		   end
		   else if(state[CHECK_SPI_COMPLETE] & spi_transmit_total_count_reg)
		   begin
			  spi_transmit_total_count_reg <= 1'b0;
		   end
		   else if(state[IDLE])
		   begin
		      master_spi_initiate_transmit <= 1'b0;
		      spi_transmit_total_count <= 0;
			  spi_transmit_total_count_reg <= 1'b0;
		   end
		end
	end

//=================================================================================
//	SPI Transmit Cycle: Load SPI Transmit Array 
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   spi_payload_byte_count <= 0;
		   spi_payload_byte_count_reg <= 1'b0;
		end
		else
		begin
		   if(state[LATCH_COMMAND])
		   begin
			  spi_payload_bytes[spi_payload_byte_count] <= spi_command_to_send;
		   end
		   else if(state[CHECK_READ_WRITE_FLAG] & spi_read_write_flag)
		   begin
			  //Special cases to handle PGA_READ_COMMAND
			     spi_payload_byte_count <= spi_payload_byte_count + 1'd1;
			     spi_payload_bytes[1] <= 8'h00;
		   end
		   else if(state[MEMORY_RD_ASSERT_OE] & !spi_payload_byte_count_reg)
		   begin
		      spi_payload_byte_count_reg <= 1'b1;
		      spi_payload_byte_count <= spi_payload_byte_count + 1'd1;
		   end
		   else if(state[MEMORY_RD_LATCH])
		   begin
		      spi_payload_byte_count_reg <= 1'b0;
			  spi_payload_bytes[spi_payload_byte_count] <= MEMORY_DATA;
		   end
		   else if(state[IDLE])
		   begin
		      spi_payload_byte_count <= 0;
		      spi_payload_byte_count_reg <= 1'b0;
		   end
		end
	end
			
//=================================================================================
//  SPI Transmit Cycle: Transmission Complete
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   SPI_TRANSMIT_COMPLETE <= 1'b0;
		end
  		else
		begin
		   if(state[CHECK_SPI_COMPLETE] & (spi_transmit_total_count >= number_of_spi_bytes_to_send))
		      SPI_TRANSMIT_COMPLETE <= 1'b1;
		   else if(state[IDLE])
		      SPI_TRANSMIT_COMPLETE <= 1'b0;
		end
   	end

	
//=================================================================================
//	Write Cycle: Memory Write Address
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_wr_address_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_WR_ADDRESS])
		      memory_wr_address_flag <= 1'b1;
		   else if(state[CHECK_SPI_COMPLETE])
		      memory_wr_address_flag <= 1'b0;
		end
	end

//=================================================================================
//	Write Cycle: Memory Write Data
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_wr_data_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_WR_DATA])
		   begin
		      memory_wr_data_flag <= 1'b1;
		   end
		   else if(state[CHECK_SPI_COMPLETE])
		      memory_wr_data_flag <= 1'b0;
		end
	end
	
//=================================================================================
//	Write Cycle: Memory Write Latch
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		   memory_wr_latch_flag <= 1'b0;
		else
		begin
		   if(state[MEMORY_WR_LATCH])
		   begin
		      memory_wr_latch_flag <= 1'b1;
		   end
		   else if(state[CHECK_SPI_COMPLETE])
		      memory_wr_latch_flag <= 1'b0;
		end
	end
			
//=================================================================================
//	SPI Receive Cycle: Start 
//=================================================================================
	always @( posedge CLK or negedge RST_N )
	begin
		if ( !RST_N )
		begin
		   master_spi_initiate_receive <= 1'b0;
		   spi_receive_total_count <= 0;
		   start_spi_receive_reg <= 1'b0;
		end
		else
		begin
		   if(state[START_SPI_TRANSMIT])
		   begin
		      spi_receive_total_count <= number_of_spi_read_bytes;
		   end
		   else if(state[START_SPI_RECEIVE] & !start_spi_receive_reg)
		   begin
		      start_spi_receive_reg <= 1'b1;
		      if(spi_receive_total_count > 0)
		         spi_receive_total_count <= spi_receive_total_count - 1'd1;
		      master_spi_initiate_receive <= 1'b1;
		   end
		   else if(state[MEMORY_WR_ADDRESS])
		   begin
		      start_spi_receive_reg <= 1'b0;
		      master_spi_initiate_receive <= 1'b0;
		   end
		end
	end


//=================================================================================
//	Serial I2C Communications Finite State Machine
//              IDLE                       = 0,
//              START                      = 1,
//              LATCH_COMMAND              = 2,
//              CHECK_READ_WRITE_FLAG      = 3,
//              MEMORY_RD_ASSERT_OE        = 4,
//              MEMORY_RD_ADDRESS          = 5,
//              MEMORY_RD_DATA             = 6,
//              MEMORY_RD_LATCH            = 7,
//              START_SPI_TRANSMIT         = 8,
//              CHECK_SPI_COMPLETE         = 9,
//              START_SPI_TRANSMIT_DELAY   = 10
//              MEMORY_WRITE_CYCLE         = 11,
//              MEMORY_WR_ADDRESS          = 12,
//              MEMORY_WR_DATA             = 13,
//              MEMORY_WR_LATCH            = 14;
//
//=================================================================================

	always @(  posedge CLK or negedge RST_N  )
	begin
		if( !RST_N )
		begin
			state <= 17'h00;
			state [IDLE] <= 1'b1;
		end
		else
			state <= next;
	end


	always @( state or SPI_COMMAND_LATCH or spi_read_write_flag
	         or GRANT_RDY or spi_payload_byte_count or master_spi_transmit_complete
			 or number_of_spi_bytes_to_send or master_spi_byte_received
			 or memory_rd_latch_flag or master_spi_initiate_receive)
	begin
		next = 17'h00;
		if(state[IDLE])
		begin
			if(SPI_COMMAND_LATCH)
				next[START] = 1'b1;
			else
				next[IDLE] = 1'b1;
		end

		if(state[START])
		begin
          next[LATCH_COMMAND] = 1'b1;
		end

		if(state[LATCH_COMMAND])
		begin
			next[CHECK_READ_WRITE_FLAG] = 1'b1;
		end

		if(state[CHECK_READ_WRITE_FLAG])
		begin
			if(!spi_read_write_flag)
				next[MEMORY_RD_ASSERT_OE] = 1'b1;
			else
				next[START_SPI_TRANSMIT] = 1'b1;
		end

		if(state[MEMORY_RD_ASSERT_OE])
		begin
			if(GRANT_RDY)
				next[MEMORY_RD_ADDRESS] = 1'b1;
			else
				next[MEMORY_RD_ASSERT_OE] = 1'b1;
		end

		if(state[MEMORY_RD_ADDRESS])
		begin
           next[MEMORY_RD_DATA] = 1'b1;
		end

		if(state[MEMORY_RD_DATA])
		begin
           next[MEMORY_RD_LATCH] = 1'b1;
		end

		if(state[MEMORY_RD_LATCH])
		begin
		   if(spi_payload_byte_count < (number_of_spi_payload_bytes))
		      next[MEMORY_RD_ASSERT_OE] = 1'b1;
		   else if(memory_rd_latch_flag)
		      next[START_SPI_TRANSMIT] = 1'b1;
		   else
		      next[MEMORY_RD_LATCH] = 1'b1;
		end

		if(state[START_SPI_TRANSMIT])
		begin
				next[WAIT_FOR_SPI_COMPLETE] = 1'b1;
		end

		if(state[WAIT_FOR_SPI_COMPLETE])
		begin
			if(master_spi_transmit_complete)
				next[CHECK_SPI_COMPLETE] = 1'b1;
			else
				next[WAIT_FOR_SPI_COMPLETE] = 1'b1;
		end

		if(state[CHECK_SPI_COMPLETE])
		begin
            if(spi_transmit_total_count < number_of_spi_bytes_to_send)
				next[START_SPI_TRANSMIT_DELAY] = 1'b1;
			else if(spi_read_write_flag)
			begin
			   if(spi_receive_total_count >= 3'h1)
			      next[START_SPI_RECEIVE] = 1'b1;
			   else
                  next[IDLE] = 1'b1;
			end
			else
				next[IDLE] = 1'b1;
		end

		if(state[START_SPI_TRANSMIT_DELAY])
		begin
			if(!master_spi_transmit_complete)
				next[START_SPI_TRANSMIT] = 1'b1;
			else
				next[START_SPI_TRANSMIT_DELAY] = 1'b1;
		end

		if(state[START_SPI_RECEIVE])
		begin
			if(master_spi_initiate_receive)
				next[MEMORY_WRITE_CYCLE] = 1'b1;
			else
				next[START_SPI_RECEIVE] = 1'b1;
		end

		if(state[MEMORY_WRITE_CYCLE])
		begin
			if(master_spi_byte_received)
			   next[MEMORY_WR_ADDRESS] = 1'b1;
			else
			   next[MEMORY_WRITE_CYCLE] = 1'b1;
		end

		if(state[MEMORY_WR_ADDRESS])
		begin
           next[MEMORY_WR_DATA] = 1'b1;
		end

		if (state[MEMORY_WR_DATA])
		begin 
           next[MEMORY_WR_LATCH] = 1'b1;
		end

		if (state[MEMORY_WR_LATCH])
		begin
	       if(!master_spi_byte_received)
              next[CHECK_SPI_COMPLETE] = 1'b1;
			else
               next[MEMORY_WR_LATCH] = 1'b1;
		end

`ifdef SIM
begin
	if ( state == (1 << IDLE))
		state_name = "IDLE";		
	if ( state == (1 << START))
		state_name = "START";		
	if ( state == (1 << LATCH_COMMAND))
		state_name = "LATCH_COMMAND";		
	if ( state == (1 << CHECK_READ_WRITE_FLAG))
		state_name = "CHECK_READ_WRITE_FLAG";		
	if ( state == (1 << MEMORY_RD_ASSERT_OE))
		state_name = "MEMORY_RD_ASSERT_OE";		
	if ( state == (1 << MEMORY_RD_ADDRESS))
		state_name = "MEMORY_RD_ADDRESS";
	if ( state == (1 << MEMORY_RD_DATA))
		state_name = "MEMORY_RD_DATA";
	if ( state == (1 << MEMORY_RD_LATCH))
		state_name = "MEMORY_RD_LATCH";
	if ( state == (1 << START_SPI_TRANSMIT))
		state_name = "START_SPI_TRANSMIT";
	if ( state == (1 << WAIT_FOR_SPI_COMPLETE))
		state_name = "WAIT_FOR_SPI_COMPLETE";
	if ( state == (1 << CHECK_SPI_COMPLETE))
		state_name = "CHECK_SPI_COMPLETE";
	if ( state == (1 << START_SPI_TRANSMIT_DELAY))
		state_name = "START_SPI_TRANSMIT_DELAY";
	if ( state == (1 << START_SPI_RECEIVE))
		state_name = "START_SPI_RECEIVE";
	if ( state == (1 << MEMORY_WRITE_CYCLE))
		state_name = "MEMORY_WRITE_CYCLE";
	if ( state == (1 << MEMORY_WR_ADDRESS))
		state_name = "MEMORY_WR_ADDRESS";
	if ( state == (1 << MEMORY_WR_DATA))
		state_name = "MEMORY_WR_DATA";
	if ( state == (1 << MEMORY_WR_LATCH))
		state_name = "MEMORY_WR_LATCH";
end
`endif		
	end

   //-----------------------------------------------
   // Instantiate the SPI Interface
   //-----------------------------------------------
	spi_iface                   SPI_IFACE_INST
	(
	.CLK                        (CLK), 
	.RST_N                      (RST_N),

	.SPI_DATA_OUT               (SPI_DATA_OUT),
	.SPI_CLK                    (SPI_CLK),
	.SPI_DATA_IN                (SPI_DATA_IN),
	.SPI_CS_N                   (spi_cs_from_spi_iface),

	.DIN                        (spi_transfer_storage_register),
	.WE_IN                      (master_spi_initiate_transmit),
	.RDY_IN                     (master_spi_transmit_complete),

	.DOUT                       (serial_receive_out_byte),
	.WE_OUT_N                   (master_spi_byte_received),
	.RDY_OUT                    (master_spi_initiate_receive)

	);



endmodule

