/******************************************************
 * Title: ad9283_adc_model.v 
 * Author: Richard Jolly                           
 * Creation Date: 2/26/2016                           
 * Description: This Module is models the AD9283 Single Channel
 * ADC. It uses the ADC_ENCODE signal as a clock to convert the
 * analog input to an eight bit digital word. The output is a 
 * parallel eight bits.
 *
 * Copyright Earth People Technology, Inc. 2016
 *
******************************************************/
`include "../src/define.v"

`timescale 1 ns / 10 ps

module ad9283_adc_model
  (
	input wire                  CLK,
	input wire                  RST_N,

    // ADC Interface	
	input wire                  ADC_ENCODE,
	output wire  [7:0]          ADC_BUS,
	
	input wire   [3:0]          ADC_CHANNEL

   );


   //----------------------------------------------
   // Parameter Declarations
   //----------------------------------------------
   parameter 	IDLE                        = 0,
				WAIT_FOR_EOC                = 1,
                ASSERT_ADC_CONTROL          = 2,
                READ_ADC                    = 3,
                ASSERT_USB_RXF              = 4,
                UPDATE_CH_SELECT            = 5,
			    DATA_BYTE_COMPLETE          = 6;
   
   parameter 	ADC_EOC_DELAY_COUNT         = 16'h7f;
   //------------------------------------------------
   // Wire and Register Declarations                
   //------------------------------------------------
   	reg [6:0]                   state, next;


	//ADC Registers
   	wire   [7:0]                adc_data_out;
   	reg                         adc_cs;
   	reg                         adc_read;
	wire                        adc_busy;
	wire                        adc_conv_start;
	reg                         adc_end_of_conv;
	reg                         adc_encode_reg;	
	reg                         adc_encode_flag;	
	
	//Data Transfer registers
   	reg   [7:0]                 transmit_data_out;
	
    // ADC EOC Registers
	reg                         adc_eoc_reg;
   	reg	[15:0]                  adc_eoc_counter;
	
	//ADC Channel Setup
	wire                        adc_single_channel;
	
	//ADC Internal Registers
   	wire   [7:0]                adc_conversion_register;
   	wire   [7:0]                adc_setup_register;	
   	wire   [7:0]                adc_averaging_register;
   	wire   [7:0]                adc_reset_register;	
	
	//Number of Channels
	reg  [2:0]                  number_of_scan_channels;
	reg                         num_chans_decrement_reg;
	

       
`ifdef SIM
   	reg [8*30:1] state_name;
`endif
   
   //------------------------------------------------
   // Signal Assignments  
   //------------------------------------------------
   
    
   assign  adc_conv_start = adc_encode_reg;
   assign  ADC_EOC = ~adc_end_of_conv;
   
   
   //Set the ADC registers in the adc_model
   assign  adc_conversion_register = {4'h0,ADC_CHANNEL};
   assign  adc_setup_register = 8'h00; 
   assign  adc_averaging_register = 8'h00; 
   assign  adc_reset_register = 8'h00;
   
   //ADC Data Bus connection
   assign ADC_BUS = adc_data_out;
   
   //ADC Channel Setup
   assign adc_single_channel = 1'b0;
   
   //-----------------------------------------------
   // Register the ADC_ENCODE signal for use in the
   // FSM
   // 
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
			adc_encode_reg <= 1'b0;
			adc_encode_flag <= 1'b0;
	end
	else 
	  begin
	     if (ADC_ENCODE & !adc_encode_flag)
		 begin
	  		 adc_encode_reg <= 1'b1;
			 adc_encode_flag <= 1'b1;
		 end
	     else if (ADC_ENCODE & adc_encode_flag)
		 begin
	  		 adc_encode_reg <= 1'b0;
			 adc_encode_flag <= 1'b1;
		 end
	     else if (!ADC_ENCODE & adc_encode_flag)
		 begin
	  		 adc_encode_reg <= 1'b0;
			 adc_encode_flag <= 1'b0;
		 end
	  end
	end

   //-----------------------------------------------
   // ADC Cycle: Set Number of channels to scan 
   //-----------------------------------------------

   always @( negedge CLK or negedge RST_N )
   begin
      if ( !RST_N )
      begin
         number_of_scan_channels <= 3'h0;
		 num_chans_decrement_reg <= 1'b0;
      end
      else
      begin
         if(state[ASSERT_ADC_CONTROL])
         begin
		    case(adc_conversion_register)
			8'h00:
               number_of_scan_channels <= 3'h1;
			8'h08:
               number_of_scan_channels <= 3'h2;
			8'h10:
               number_of_scan_channels <= 3'h3;
			8'h18:
               number_of_scan_channels <= 3'h4;
			default:
               number_of_scan_channels <= 3'h1;
			endcase
         end
         else if(state[UPDATE_CH_SELECT] & !num_chans_decrement_reg)
         begin
            number_of_scan_channels <= number_of_scan_channels - 1'd1;
			num_chans_decrement_reg <= 1'b1;
         end
		 else if(state[DATA_BYTE_COMPLETE] & num_chans_decrement_reg)
         begin
			num_chans_decrement_reg <= 1'b0;
         end
      end
   end

   //-----------------------------------------------
   // Assert the ADC_EOC signal to signal the ADC
   // has completed the conversion and data is ready 
   // for transmission.
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	  		adc_end_of_conv <= 1'b0;
			adc_eoc_reg <= 1'b0;
			adc_eoc_counter <= 0;
	end
	else 
	  begin
	     if (state[ASSERT_ADC_CONTROL] & !adc_eoc_reg)
		 begin
	  		 adc_end_of_conv <= 1'b1;
			 adc_eoc_reg <= 1'b1;
		 end
	     else if (adc_eoc_reg)
		 begin
		     if(adc_eoc_counter < ADC_EOC_DELAY_COUNT)
	  		    adc_eoc_counter <= adc_eoc_counter + 1'd1;
			 else 
			 begin
	  		    adc_end_of_conv <= 1'b0;
			    adc_eoc_reg <= 1'b0;
			    adc_eoc_counter <= 0;
			 end
		 end
	  end
	end


   //-----------------------------------------------
   // Read the ADC Data.
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	  		transmit_data_out <= 0;
	end
	else 
	  begin
	     if (state[ASSERT_USB_RXF])
		 begin
	  		 transmit_data_out <= adc_data_out;
		 end
	  end
	end

	//-----------------------------------------------
   // Assert ADC controls to read the ADC Data.
   //-----------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	  		adc_cs <= 1'b1;
		  	adc_read <= 1'b1;
	end
	else 
	  begin
	     if (state[ASSERT_USB_RXF])
		 begin
	  		   adc_cs <= 1'b0;
		  	   adc_read <= 1'b0;
		 end
	     else if (state[UPDATE_CH_SELECT])
		 begin
	  		adc_cs <= 1'b1;
		  	adc_read <= 1'b1;
		 end
	  end
	end

   //--------------------------------------------------
   // Write State Machine
   //--------------------------------------------------  
   // Next State Logic
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  begin
	     state <= 7'h000;
	     state[IDLE] <= 1'b1;
	  end
	else
	  state <= next;
     end


   always @ ( state or adc_busy or adc_encode_reg
	      or adc_conv_start 
	      or adc_cs or adc_read)

     begin
	next = 7'h000;

	if (state[IDLE])
	  begin
	     if(adc_encode_reg)
	       next[WAIT_FOR_EOC] = 1'b1;
		else
	       next[IDLE] = 1'b1;			
	  end

	if (state[WAIT_FOR_EOC])
	begin
	    if(adc_busy)
		    next[ASSERT_ADC_CONTROL] = 1'b1;
	    else 
		    next[WAIT_FOR_EOC] = 1'b1;
	end

	if (state[ASSERT_ADC_CONTROL])
	begin
	    if(!adc_busy)
		    next[READ_ADC] = 1'b1;
	    else 
		    next[ASSERT_ADC_CONTROL] = 1'b1;
	end


	if (state[READ_ADC])
	  next[ASSERT_USB_RXF] = 1'b1;

	if (state[ASSERT_USB_RXF])
	       next[UPDATE_CH_SELECT] = 1'b1;

	if (state[UPDATE_CH_SELECT])
	  begin
	  	if(!adc_encode_reg)
			next[DATA_BYTE_COMPLETE] = 1'b1;
		else
			next[UPDATE_CH_SELECT] = 1'b1;			
	  end 

	if (state[DATA_BYTE_COMPLETE])
	begin
		if(!adc_encode_reg)
			next[IDLE] = 1'b1;			
		else 
			next[DATA_BYTE_COMPLETE] = 1'b1;	
	end
	
`ifdef SIM
	if (state == (1 << IDLE))
	  state_name = "IDLE";
	else if (state == (1 << WAIT_FOR_EOC))
	  state_name = "WAIT_FOR_EOC";
	else if (state == (1 << ASSERT_ADC_CONTROL))
	  state_name = "ASSERT_ADC_CONTROL";
	else if (state == (1 << READ_ADC))
	  state_name = "READ_ADC";
	else if (state == (1 << ASSERT_USB_RXF))
	  state_name = "ASSERT_USB_RXF";
	else if (state == (1 << UPDATE_CH_SELECT))
	  state_name = "UPDATE_CH_SELECT";
	else if (state == (1 << DATA_BYTE_COMPLETE))
	  state_name = "DATA_BYTE_COMPLETE";

	  
`endif
     end // always @ (...

   adc_model                    ADC_MODEL_INST
     (
      .CLK                      (CLK), 
      .RESET                    (RST_N),

      .AD_DATA                  (adc_data_out),
      .AD_BUSY                  (adc_busy),  
      .AD_REFOUT                (),

      .AD_CSL                   (adc_cs),
      .AD_RDL                   (adc_read),
      .AD_CONVST                (adc_conv_start),
	  
	  .ADC_MULTI_CHANNEL        (adc_single_channel),
	  
	  .ADC_CONVERSION_REG       (adc_conversion_register),
	  .ADC_SETUP_REG            (adc_setup_register),
	  .ADC_AVERAGING_REG        (adc_averaging_register),
	  .ADC_RESET_REG            (adc_reset_register)

      );

endmodule // audio_SPI_fsm

   