//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2016
//#
//#
//# File Name:  i2c_ack.v
//# Author:     Earth People Technology
//# Date:       July 26, 2016
//# Revision:   A
//#
//# Development: EPT I2C Slave Project 
//# Application: Altera Cyclone IV FPGA
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. 
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			7/26/16 	1			Created
//#
//#									
//#
//#######################################################################
//`include "fpga_defs.v"

`resetall
`timescale 1ns/10ps
module i2c_ack_model
  (
    input  wire                 CLK,
    input  wire                 RST_N,
	
    input  wire                 GNT_START,
    input  wire                 REQ_START,
	input  wire                 RECEIVE_UNIT_DONE,
	input  wire                 TRANSMIT_UNIT_DONE,
	output  wire                VALID_SEND,
	output  reg                 VALID_RECEIVE,

    input  wire                 DATA_IN,   
    output  reg                 DOUT,
    output  wire                DATA_OUT,
    output  reg                 DONE,
    input   wire                SCLK ,
	
	input wire                  I2C_STOP,
	
	output wire [3:0]           TEST_BUS_OUT,
	
    output reg   [3:0]          STATE_OUT
   
	
   );

   //----------------------------------------------
   // Parameter Declarations
   //----------------------------------------------
   parameter 	IDLE                     = 0,
			RECEIVE_ACK			         = 1,
			READ_ACK                     = 2,
            RECEIVE_ACK_DELAY            = 3,
			RECEIVE_ACK_COMPLETE         = 4,
			SEND_ACK                     = 5,
			WRITE_ACK                    = 6,
			SEND_ACK_COMPLETE            = 7,
			SEND_ACK_DELAY               = 8;
			
`ifdef SIM
   parameter    SEND_ACK_DELAY_COUNT     = 16'h60;
   parameter    SEND_ACK_COMPLETE_COUNT  = 16'h60;
`elsif
   parameter    SEND_ACK_DELAY_COUNT     = 16'h160;
   parameter    SEND_ACK_COMPLETE_COUNT  = 16'h160;
`endif   
   //------------------------------------------------
   // Wire and Register Declarations                
   //------------------------------------------------
	reg	[3:0]					bit_count;
	reg [1:0]                   bit_count_reg;
  
	reg	[9:0]					state, next;

	// sclk_reg to determine value of the output SCLK
	wire	[11:0]				sclk_reg;

	// sdata_reg_reg to determine value of the output DATA_OUT
	wire	[11:0]				sdata_reg;
	
	//DOUT Registers
	reg                         dout_reg;
	
	//SCL sync registers
	reg                         meta_i2c_clock;
    reg                         I2C_SCL;
	
	//State machine counters
    reg                         send_ack_delay_flag;
    reg  [15:0]                 send_ack_delay_counter;
    reg                         send_ack_complete_flag;
    reg  [15:0]                 send_ack_complete_counter;	

`ifdef SIM
   reg [8*20:1] state_name;
`endif

//------------------------------------------------
//	Assignments   
//------------------------------------------------

//	assign	sdata_reg = {state[SEND_ACK] , state[WRITE_ACK] , state[SEND_ACK_LO] , 
//	                     state[SEND_ACK_HI] , state[SEND_ACK_DELAY] , state[SEND_ACK_COMPLETE] ,
//						 state[IDLE]};

    assign  DATA_OUT = VALID_RECEIVE ? 1'b1 : 1'b0;
	assign  VALID_SEND = (state[SEND_ACK_COMPLETE] | state[SEND_ACK_DELAY]) ? 1'b1 : 1'b0;
						 
						 
    //TEST ONLY
    assign TEST_BUS_OUT[3:0] = 	bit_count;

   //-----------------------------------------------
   // Create a State Test Output Signal
   //-----------------------------------------------
	always @(posedge CLK or negedge RST_N)
 	begin
      if (!RST_N)
	  begin
	     STATE_OUT <= 0;
	  end
	else
	  begin
	     if (state[IDLE])
               STATE_OUT <= 5'h0;
	     if (state[RECEIVE_ACK])
	       STATE_OUT <= 5'h1;
	     if (state[READ_ACK])
	       STATE_OUT <= 5'h2;
	     if (state[RECEIVE_ACK_DELAY])
	       STATE_OUT <= 5'h3;
	     if (state[RECEIVE_ACK_COMPLETE])
	       STATE_OUT <= 5'h4;
	     if (state[SEND_ACK])
	       STATE_OUT <= 5'h5;
	     if (state[WRITE_ACK])
	       STATE_OUT <= 5'h6;
	     if (state[SEND_ACK_DELAY])
	       STATE_OUT <= 5'h7;
	     if (state[SEND_ACK_COMPLETE])
	       STATE_OUT <= 5'h8;
	  end // else: !if(!RST_N)
     end // always @ (posedge CLK or negedge RST_N)

//=================================================================================
//	Send I2C_Clock to sync registers
//=================================================================================

	always @( negedge CLK or negedge RST_N )
	begin
         meta_i2c_clock <= SCLK;
		 I2C_SCL <= meta_i2c_clock;
	end

   //------------------------------------------------
   // Timer to exit the state[SEND_ACK_DELAY]
   //------------------------------------------------
	always @(posedge CLK or negedge RST_N)
	begin
    if (!RST_N)
    begin
            send_ack_delay_flag <= 1'b0;
			send_ack_delay_counter <= 0;
	end
	else
	begin
		if (state[SEND_ACK_DELAY])
		begin
			if(send_ack_delay_counter < SEND_ACK_DELAY_COUNT)
			   send_ack_delay_counter <= send_ack_delay_counter + 1'd1;
			else
            begin
			   send_ack_delay_flag <= 1'b1;
			   send_ack_delay_counter <= 0;
            end			
	    end
	    else if(state[SEND_ACK_COMPLETE])
        begin
            send_ack_delay_flag <= 1'b0;
			send_ack_delay_counter <= 0;
	    end
	end
	end // always @ (posedge CLK or negedge RST_N)

   //------------------------------------------------
   // Timer to exit the state[SEND_ACK_COMPLETE]
   //------------------------------------------------
	always @(posedge CLK or negedge RST_N)
	begin
    if (!RST_N)
    begin
            send_ack_complete_flag <= 1'b0;
			send_ack_complete_counter <= 0;
	end
	else
	begin
		if (state[SEND_ACK_COMPLETE])
		begin
			if(send_ack_complete_counter < SEND_ACK_COMPLETE_COUNT)
			   send_ack_complete_counter <= send_ack_complete_counter + 1'd1;
			else
            begin
			   send_ack_complete_flag <= 1'b1;
			   send_ack_complete_counter <= 0;
            end			
	    end
	    else if(state[IDLE])
        begin
            send_ack_complete_flag <= 1'b0;
			send_ack_complete_counter <= 0;
	    end
	end
	end // always @ (posedge CLK or negedge RST_N)

   //------------------------------------------------
   // VALID_REVEIVE register to indicate when the ack
   // unit is checking the received ack bit
   //------------------------------------------------
	always @(posedge CLK or negedge RST_N)
	begin
		if (!RST_N)
			VALID_RECEIVE <= 0;
	else
	begin
		if ( state[READ_ACK] || state[RECEIVE_ACK_COMPLETE] )
			VALID_RECEIVE <= 1;
	  else
	     VALID_RECEIVE <= 0;
	end
	end // always @ (posedge CLK or negedge RST_N)

	
   //------------------------------------------------
   // Create a DONE Signal when the byte has been transmitted
   //------------------------------------------------
	always @(posedge CLK or negedge RST_N)
	begin
		if (!RST_N)
			DONE <= 0;
	else
	begin
	  if ( state[SEND_ACK_COMPLETE] | state[RECEIVE_ACK_COMPLETE] )
         DONE <= 1;
	  else
	     DONE <= 0;
	end
	end // always @ (posedge CLK or negedge RST_N)

//------------------------------------------------
// Create a bit_count of the number of bits sent
//I2C_SCL
//------------------------------------------------
/*   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
       bit_count <= 0;
       bit_count_reg <= 2'b00;
	end
	else
	  begin
	     if((state[RECEIVE_ACK]) & (bit_count_reg == 2'b00))
		 begin
		   bit_count_reg <= 2'b11;
		 end
	     //else if(I2C_SCL & (bit_count_reg == 2'b01))
	     else if((state[RECEIVE_ACK_HI] | state[RECEIVE_ACK_COMPLETE]) & (bit_count_reg == 2'b01))
		 begin
	       bit_count <= bit_count + 1'd1;
		   bit_count_reg <= 2'b11;
		 end
	     //else if(!I2C_SCL & (bit_count_reg == 2'b11))
	     else if((state[RECEIVE_ACK_LO] | state[READ_ACK]) & (bit_count_reg == 2'b11))
		   bit_count_reg <= 2'b01;
		 //else if(state[RECEIVE_ACK_COMPLETE] | state[SEND_ACK_COMPLETE] | state[IDLE])
		 else if(bit_count >= 8'h9)
		 begin
	       bit_count <= 0;
		   bit_count_reg <= 2'b00;
		 end
	  end
     end 
 */ 
   //------------------------------------------------
   // Read Ack Bit
   //------------------------------------------------
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	begin
	  DOUT <= 1'b0;
	  dout_reg <= 1'b0;
	end
	else
	  begin
	     if(state[RECEIVE_ACK_COMPLETE] & !dout_reg)
         begin
             DOUT <= DATA_IN;
			 dout_reg <= 1'b1;
         end
	     else if(REQ_START & dout_reg)
		 begin
	        DOUT <= 1'b0;
            dout_reg <= 1'b0;
		 end
	  end // else: !if(!RST_N)
     end // always @ (posedge CLK or negedge RST_N)     
	  
//--------------------------------------------------
// Finite State Machine
//
//parameter 	IDLE                     = 0,
//			RECEIVE_ACK                  = 1,
//			READ_ACK                     = 2,
//          RECEIVE_ACK_DELAY            = 3,
//			RECEIVE_ACK_COMPLETE         = 4,
//			SEND_ACK                     = 5,
//			WRITE_ACK                    = 6,
//          SEND_ACK_DELAY               = 7,
//			SEND_ACK_COMPLETE            = 8;
//
//--------------------------------------------------  
// Next State Logic
   always @(posedge CLK or negedge RST_N)
     begin
	if (!RST_N)
	  begin
	     state <= 12'h00;
	     state[IDLE] <= 1'b1;
	  end
	else
	  state <= next;
     end

   always @ ( state or DATA_IN or bit_count or
			GNT_START or REQ_START or I2C_SCL or
			RECEIVE_UNIT_DONE or I2C_STOP or 
			send_ack_delay_flag or send_ack_complete_flag or 
			TRANSMIT_UNIT_DONE)
     begin
	next = 12'h00;

	if (state[IDLE])
	  begin
		if ( REQ_START )
			next[RECEIVE_ACK] = 1'b1;
		else if ( GNT_START )
			next[SEND_ACK] = 1'b1;
		else
			next[IDLE] = 1'b1;
	  end

	if (state[RECEIVE_ACK])
	begin
	   if(TRANSMIT_UNIT_DONE)
	      next[READ_ACK] = 1'b1;
       else
	      next[RECEIVE_ACK] = 1'b1;
	end

	if (state[READ_ACK])
	begin
		if (!TRANSMIT_UNIT_DONE)
			next[RECEIVE_ACK_COMPLETE] = 1'b1;
	     else
			next[READ_ACK] = 1'b1;
	end

	if (state[RECEIVE_ACK_DELAY])
	begin
		if(I2C_SCL)
			next[RECEIVE_ACK_COMPLETE] = 1'b1;
	    else if(I2C_STOP)
		    next[IDLE] = 1'b1;
		else
			next[RECEIVE_ACK_DELAY] = 1'b1;			
	end
	
	if (state[RECEIVE_ACK_COMPLETE])
	begin
		if(!I2C_SCL)
		   next[IDLE] = 1'b1;
		else
		   next[RECEIVE_ACK_COMPLETE] = 1'b1;			
	end

	if (state[SEND_ACK])
	begin
	   if(RECEIVE_UNIT_DONE)
          next[WRITE_ACK] = 1'b1;
       else
          next[SEND_ACK] = 1'b1;
	end

	if (state[WRITE_ACK])
	begin
	   if(!RECEIVE_UNIT_DONE)
          next[SEND_ACK_DELAY] = 1'b1;	
       else
	      next[WRITE_ACK] = 1'b1;
	end

	if (state[SEND_ACK_DELAY])
	begin
		//if(I2C_SCL)
		if(send_ack_delay_flag)
			next[SEND_ACK_COMPLETE] = 1'b1;
	    else if(I2C_STOP)
		    next[IDLE] = 1'b1;
		else
			next[SEND_ACK_DELAY] = 1'b1;			
	end
	
	if (state[SEND_ACK_COMPLETE])
	begin
		//if(!I2C_SCL)
		if(send_ack_complete_flag)
			next[IDLE] = 1'b1;
	    else if(I2C_STOP)
		    next[IDLE] = 1'b1;
		else
			next[SEND_ACK_COMPLETE] = 1'b1;			
	end


`ifdef SIM
	if (state == (1 << IDLE))
	  state_name = "IDLE";
	else if (state == (1 << RECEIVE_ACK))
	  state_name = "RECEIVE_ACK";
	else if (state == (1 << READ_ACK))
	  state_name = "READ_ACK";
	else if (state == (1 << RECEIVE_ACK_DELAY))
	  state_name = "RECEIVE_ACK_COMPLETE";	
	else if (state == (1 << RECEIVE_ACK_COMPLETE))
	  state_name = "RECEIVE_ACK_COMPLETE";	
	else if (state == (1 << SEND_ACK))
	  state_name = "SEND_ACK";	
	else if (state == (1 << WRITE_ACK))
	  state_name = "WRITE_ACK";
	else if (state == (1 << SEND_ACK_DELAY))
	  state_name = "SEND_ACK_DELAY";	
	else if (state == (1 << SEND_ACK_COMPLETE))
	  state_name = "SEND_ACK_COMPLETE";	
`endif
     end // always @ (...

   
endmodule 


   