//#######################################################################
//#
//#	Copyright 	Earth People Technology Inc. 2014
//#
//#
//# File Name:  i2c_comms.v
//# Author:     Earth People Technology
//# Date:       November 26, 2014
//# Revision:   A
//#
//# Development: EPT I2C Slave Project 
//# Application: Altera Cyclone IV FPGA
//# Description: This file contains verilog code which will allow access
//#              to Active Transfer Library. 
//#               
//#              
//#
//#************************************************************************
//#
//# Revision History:	
//#			DATE		VERSION		DETAILS		
//#			11/26/14 	1			Created
//#
//#									
//#
//#######################################################################

module i2c_slave_comms(
	input  wire                 CLK,
	input  wire                 RESET,

	inout  wire                 IIC_DATA,
	input  wire                 IIC_CLOCK,

	input  wire [7:0]           SERIAL_IIC_TRANSMIT_DATA_IN,
	output reg                  SERIAL_IIC_TRANSMIT_ENA_OUT,
	input  wire                 SERIAL_IIC_TRANSMIT_RDY_IN,

	output wire [7:0]           SERIAL_IIC_RECEIVE_DATA_OUT,
	input  wire                 SERIAL_IIC_RECEIVE_ENA_IN,
	output  reg                 SERIAL_IIC_RECEIVE_RDY_OUT,
	
	output reg  [7:0]           ERROR_OUT,
	
	output wire [11:0]           TEST_BUS_OUT,
    output wire  [3:0]          STATE_OUT


		);

//==================================================================
// Parameter Declarations
//==================================================================
	parameter	IDLE                         = 0,
                I2C_START                    = 1,
                READ_ADDRESS                 = 2,
                DECODE_RW                    = 3,
				CHECK_READ_WRITE             = 4,
                WRITE_BYTE                   = 5,
                WRITE_ACK                    = 6,
                WRITE_ACK_DELAY              = 7,
                READ_BYTE                    = 8,
                READ_ACK                     = 9,
                READ_ACK_DELAY               = 10,
                I2C_STOP                     = 11;
				
	parameter	I2C_IDLE_STATE               = 0,
                I2C_DATA_LOW_STATE           = 1,
                I2C_CLOCK_LOW_STATE          = 2,
                I2C_START_DETECT_STATE       = 3,
				I2C_ACK_DONE_STATE           = 4,
                I2C_CLOCK_HIGH_STATE         = 5,
                I2C_DATA_HIGH_STATE          = 6,
                I2C_WRITE_CLOCK_LOW_STATE    = 7, 
                I2C_WRITE_DATA_LOW_STATE     = 8,
                I2C_STOP_DETECT_STATE        = 9;
				
				
	parameter   I2C_START_STOP_COUNT        = 16'h42;
	parameter   I2C_ERROR_COUNT             = 16'h157;
	parameter   I2C_WAIT_BYTE_COUNT         = 16'h155;
	
	parameter   DEVICE_ADDRESS              = 8'h44;


//==================================================================
// Register Declarations
//==================================================================
	// State Machine Registers
    reg  [11:0]                 state;
    reg  [11:0]                 next;
	
	//Clock meta stability registers
    reg                         meta_i2c_clock;
    reg                         I2C_CLOCK;


	// Receive FIFO Storage
    reg                         receive_fifo_rden;
    reg                         receive_fifo_rden_reg;
    wire	[7:0]				receive_fifo_din;
    reg                         receive_fifo_wren;
    reg                         receive_fifo_wren_reg;
    wire	[7:0]				receive_fifo_dout;
    wire                        receive_fifo_full;
    wire                        receive_fifo_empty;
    wire                        receive_data_count;
    wire                        receive_fifo_wr_ack;
    wire                        receive_fifo_afull;
    wire                        receive_fifo_aempty;
    wire                        receive_fifo_flush;

	// Transmit FIFO Storage
    reg                         transmit_fifo_rden;
    reg   [7:0]                 transmit_fifo_din;
    reg                         transmit_fifo_wren;
    reg                         transmit_fifo_wren_reg;
    wire  [7:0]                 transmit_fifo_dout;
    wire                        transmit_fifo_full;
    wire                        transmit_fifo_empty;
    wire                        transmit_data_count;
    wire                        transmit_fifo_wr_ack;
    wire                        transmit_fifo_afull;
    wire                        transmit_fifo_aempty;
    wire                        transmit_fifo_flush;
	
	// Read Cycle Status Messages
	reg  [1:0]                  serial_iic_rdy_out_reg;

	// Write Cycle messages
  	reg                         serial_iic_ena_out_reg;

	wire	[7:0]				receive_data_out;
	wire                        receive_valid_out;
	reg                         receive_active;
	reg     [1:0]               receive_active_reg;

	// IIC Transmit unit Registers
	wire                        iic_data_from_transmit;
	wire                        iic_clock_from_transmit;
	wire	[7:0]               transmit_data_in;
	reg                         transmit_data_valid;
	reg                         transmit_active;
	wire                        transmit_done;


 	// Write cycle transmit registers
	reg                         transmission_active;
	reg   [1:0]                 transmit_data_valid_reg;
	reg                         transmit_data_complete;

  // Storing words to Receive FIFO
	reg                         wait_for_data_valid;

	// Registers to initiate the store receive unit words: receive_decode_message receive_decode_message_reg			
	reg                         receive_decode_message;
	reg                         receive_decode_message_reg;
	
	//I2C Check for next byte
	reg   [15:0]                i2c_next_byte_timeout_cntr;
	reg                         i2c_next_byte_flag;
	
	//I2C Address registers
	reg                         i2c_address_detected;
	wire  [7:0]                 i2c_address;
	
	//I2C Read Write Bit Registers
	reg                         i2c_read_write_bit;
	
	//I2C Stop Bit Registers
	reg                         i2c_rw_stop;
	reg  [1:0]                  rstop_phase;
	
	//I2C Ack 
	wire                        ack_done;
	reg                         ack_grant_start;
	reg                         ack_grant_start_reg;
	wire                        ack_request_start;
	wire                        ack_data_out;
	wire                        i2c_ack_transmit;
	wire                        i2c_data_out_from_ack;
	wire                        i2c_ack_send_active;
	wire                        i2c_ack_rcv_active;
	
	//I2C Detect
	reg                         i2c_start_detect;
	reg                         i2c_start_detect_reg;
	reg  [15:0]                  i2c_start_stop_counter;
	reg  [3:0]                  i2c_start_stop_state;
	
	//Error detection registers
	reg                         error_detect_reg;
	
	//TEST ONLY
	wire                        i2c_data_line_in;
	wire  [3:0]                 i2c_receive_out_phase;
	wire [3:0]                  i2c_ack_state;
	wire [3:0]                  i2c_ack_test_bus_out;


//==================================================================
// Simulation Directives
//==================================================================

`ifdef SIM
	reg [8*30:1] state_name;
`endif

//=================================================================================
// Assignments	
//=================================================================================
	
	assign	IIC_DATA = transmission_active ? iic_data_from_transmit : (i2c_ack_send_active ? i2c_data_out_from_ack : 1'bz);

	//I2C Address Registers
	assign  i2c_address = state[READ_ADDRESS] ? receive_data_out : 8'b0;
	
	//Receive output Registers
	assign	SERIAL_IIC_RECEIVE_DATA_OUT = receive_fifo_dout;
 
    //Transmit FIFO Registers
	assign	transmit_data_in = transmit_fifo_dout;
	
	//Receive FIFO Registers
	assign receive_fifo_din = receive_data_out;

   //I2C Ack Unit registers
	assign	ack_request_start = transmit_active;
	assign	i2c_ack_transmit = IIC_DATA;
	assign	i2c_ack_clock = I2C_CLOCK;
	
	assign  i2c_data_line_in = IIC_DATA ? 1'b1 : 1'b0;
	
	//TEST ONLY
	assign TEST_BUS_OUT[0] = I2C_CLOCK;                //8
	assign TEST_BUS_OUT[1] = IIC_DATA;                 //9
	assign TEST_BUS_OUT[2] = receive_active;           //10
	assign TEST_BUS_OUT[3] = i2c_start_detect;     //11
	assign TEST_BUS_OUT[4] = ack_request_start;//receive_valid_out;  //i2c_receive_out_phase[0];       //12
	assign TEST_BUS_OUT[5] = ack_grant_start;    //i2c_receive_out_phase[1];         //13
	assign TEST_BUS_OUT[6] = i2c_rw_stop;  //i2c_receive_out_phase[2];       //14
	assign TEST_BUS_OUT[7] = ack_done;           //i2c_receive_out_phase[3];                //15
	//assign TEST_BUS_OUT[4] = i2c_receive_out_phase[0];       //12
	//assign TEST_BUS_OUT[5] = i2c_receive_out_phase[1];         //13
	//assign TEST_BUS_OUT[6] = i2c_receive_out_phase[2];       //14
	//assign TEST_BUS_OUT[7] = i2c_receive_out_phase[3];                //15

	   assign            STATE_OUT = i2c_ack_state;
	   //assign  TEST_BUS_OUT[11:8] = i2c_start_stop_state;
	   assign  TEST_BUS_OUT[10:8] = i2c_ack_test_bus_out;
	   assign  TEST_BUS_OUT[11] = i2c_next_byte_flag;

   //-----------------------------------------------
   // Create a State Test Output Signal
   //-----------------------------------------------
/*   always @(posedge CLK or negedge RESET)
     begin
	if (!RESET)
	  STATE_OUT <= 0;
	else
	  begin
	     if (state[IDLE])
           STATE_OUT <= 5'h0;
	     if (state[I2C_START])
	       STATE_OUT <= 5'h1;
	     if (state[READ_ADDRESS])
	       STATE_OUT <= 5'h2;
	     if (state[DECODE_RW])
	       STATE_OUT <= 5'h3;
	     if (state[CHECK_READ_WRITE])
	       STATE_OUT <= 5'h4;
	     if (state[WRITE_BYTE])
	       STATE_OUT <= 5'h5;
	     if (state[WRITE_ACK])
	       STATE_OUT <= 5'h6;
	     if (state[WRITE_ACK_DELAY])
	       STATE_OUT <= 5'h7;
	     if (state[READ_BYTE])
	       STATE_OUT <= 5'h8;
	     if (state[READ_ACK])
	       STATE_OUT <= 5'h9;
	     if (state[READ_ACK_DELAY])
	       STATE_OUT <= 5'hA;
	     if (state[I2C_STOP])
	       STATE_OUT <= 5'hB;
	  end // else: !if(!RST_N)
     end // always @ (posedge CLK or negedge RST_N)
*/
//=================================================================================
//	Send I2C_Clock to sync registers
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
         meta_i2c_clock <= IIC_CLOCK;
		 I2C_CLOCK <= meta_i2c_clock;
	end

	
//=================================================================================
//	ERROR Collection: Register to capture any errors of I2C cycle
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			ERROR_OUT <= 0;			
			error_detect_reg <= 1'b0;
		end
		else
		begin
		    if(i2c_start_detect & !error_detect_reg)
			begin
			   ERROR_OUT <= 0;			
			   error_detect_reg <= 1'b1;
			end
		    else if(state[WRITE_ACK_DELAY])
			   ERROR_OUT[0] <= !ack_data_out;
			else if(i2c_rw_stop)
			   error_detect_reg <= 1'b0;
		end
	end
	
//=================================================================================
//	I2C Start Cycle: Register to indicate start of I2C cycle
//=================================================================================
    always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			i2c_start_detect <= 1'b0;			
			i2c_start_stop_counter <= 0;
			i2c_rw_stop <= 1'b0;
			i2c_start_stop_state <= I2C_IDLE_STATE;
		end
		else
		begin
		case(i2c_start_stop_state)
		I2C_IDLE_STATE:
		begin
			i2c_start_detect <= 1'b0;			
			i2c_rw_stop <= 1'b0;
			i2c_start_stop_counter <= 0;
		   if(!IIC_DATA & I2C_CLOCK)
		      i2c_start_stop_state <= I2C_DATA_LOW_STATE;
	    end
		I2C_DATA_LOW_STATE:
		   if(!IIC_DATA & !I2C_CLOCK)
		      i2c_start_stop_state <= I2C_CLOCK_LOW_STATE;
		I2C_CLOCK_LOW_STATE:
		begin
		   if(!IIC_DATA & !I2C_CLOCK)
		   begin
		      if(i2c_start_stop_counter < I2C_START_STOP_COUNT)
			     i2c_start_stop_counter <= i2c_start_stop_counter + 1'd1;
			  else
			  begin
		         i2c_start_stop_state <= I2C_START_DETECT_STATE;
				 i2c_start_stop_counter <= 0;
			  end
		   end
		   else 
		   i2c_start_stop_state <= I2C_IDLE_STATE;
		end
		I2C_START_DETECT_STATE:
		begin
		   i2c_start_detect <= 1'b1;
		   if(ack_done)
               i2c_start_stop_state <= I2C_ACK_DONE_STATE;
	    end
		I2C_ACK_DONE_STATE:
		begin
		   if(state[READ_ACK_DELAY])
		   begin
			if(!IIC_DATA & !I2C_CLOCK)
               i2c_start_stop_state <= I2C_CLOCK_HIGH_STATE;
			else if(IIC_DATA & !I2C_CLOCK)
               i2c_start_stop_state <= I2C_START_DETECT_STATE;
			end
			else if(state[WRITE_ACK_DELAY])
			   i2c_start_stop_state <= I2C_WRITE_CLOCK_LOW_STATE;
	    end
		I2C_CLOCK_HIGH_STATE:
		   if(i2c_read_write_bit & I2C_CLOCK)
               i2c_start_stop_state <= I2C_START_DETECT_STATE;
		   else if(!IIC_DATA & I2C_CLOCK)
		      i2c_start_stop_state <= I2C_DATA_HIGH_STATE;
			else if(IIC_DATA & !I2C_CLOCK)
               i2c_start_stop_state <= I2C_START_DETECT_STATE;
			else
			   i2c_start_stop_state <= I2C_CLOCK_HIGH_STATE;
		I2C_DATA_HIGH_STATE:
		begin
		   if(IIC_DATA & I2C_CLOCK)
		   begin
		         i2c_start_stop_state <= I2C_STOP_DETECT_STATE;
		   end
		   else if(!I2C_CLOCK)
		      i2c_start_stop_state <= I2C_START_DETECT_STATE;
		end
		I2C_WRITE_CLOCK_LOW_STATE:
		   if(!IIC_DATA & I2C_CLOCK)
		      i2c_start_stop_state <= I2C_WRITE_DATA_LOW_STATE;
			else if(IIC_DATA & !I2C_CLOCK)
               i2c_start_stop_state <= I2C_START_DETECT_STATE;
			else
			   i2c_start_stop_state <= I2C_WRITE_CLOCK_LOW_STATE;
		I2C_WRITE_DATA_LOW_STATE: 
		begin
		   if(IIC_DATA & I2C_CLOCK)
		   begin
		         i2c_start_stop_state <= I2C_STOP_DETECT_STATE;
		   end
		   else if(!I2C_CLOCK)
		      i2c_start_stop_state <= I2C_START_DETECT_STATE;
		end
		I2C_STOP_DETECT_STATE:
		begin
			i2c_start_detect <= 1'b0;			
			i2c_rw_stop <= 1'b1;
		   i2c_start_stop_state <= I2C_IDLE_STATE;
		end
		endcase
		end
	end

//=================================================================================
//	Check for Next Byte
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if( !RESET )
		begin
			i2c_next_byte_flag <= 1'b0;
			i2c_next_byte_timeout_cntr <= 0;
		end
		else
		begin
		   if(state[CHECK_READ_WRITE] & i2c_read_write_bit)
		   begin
			      i2c_next_byte_flag <= 1'b1;
			      i2c_next_byte_timeout_cntr <= 0;
		   end
		   else if(state[CHECK_READ_WRITE] & IIC_DATA & !I2C_CLOCK)
		   begin
			      i2c_next_byte_flag <= 1'b1;
			      i2c_next_byte_timeout_cntr <= 0;
		   end
		   else if(state[CHECK_READ_WRITE] & !IIC_DATA & I2C_CLOCK)
		   begin
			   if(i2c_next_byte_timeout_cntr < I2C_WAIT_BYTE_COUNT)
			       i2c_next_byte_timeout_cntr = i2c_next_byte_timeout_cntr + 1'd1;
				else
				begin
			       i2c_next_byte_flag <= 1'b1;
				   i2c_next_byte_timeout_cntr = 0;
				end
		   end
		   else if(state[WRITE_ACK_DELAY] | state[READ_ACK_DELAY] | state[I2C_STOP] | state[IDLE])
		   begin
			  i2c_next_byte_flag <= 1'b0;
			  i2c_next_byte_timeout_cntr <= 0;
           end		   
		end
	end

//=================================================================================
//	Write Cycle: transmission_active register
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if( !RESET )
		begin
			transmission_active <= 1'b0;
		end
		else
		begin
			if(transmit_active)
				transmission_active <= 1'b1;
			else if(i2c_ack_rcv_active)
			begin
                transmission_active <= 1'b0;
			end
		end
	end
	
	
//=================================================================================
//	Write Cycle: Store incoming word into the transmit FIFO
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
		    transmit_fifo_din <= 0;
            SERIAL_IIC_TRANSMIT_ENA_OUT <= 1'b0;
			transmit_fifo_wren <= 1'b0;			
			transmit_fifo_wren_reg <= 1'b0;
		end
		else
		begin
			if( SERIAL_IIC_TRANSMIT_RDY_IN & !transmit_fifo_wren_reg )
			begin
			    transmit_fifo_din <= SERIAL_IIC_TRANSMIT_DATA_IN;
				SERIAL_IIC_TRANSMIT_ENA_OUT <= 1'b1;
				transmit_fifo_wren <= 1'b1;
				transmit_fifo_wren_reg <= 1'b1;
			end
			if( SERIAL_IIC_TRANSMIT_RDY_IN & transmit_fifo_wren_reg )
			begin
				SERIAL_IIC_TRANSMIT_ENA_OUT <= 1'b0;
				transmit_fifo_wren <= 1'b0;
				transmit_fifo_wren_reg <= 1'b1;
			end
			else if( !SERIAL_IIC_TRANSMIT_RDY_IN & transmit_fifo_wren_reg )
			begin
				SERIAL_IIC_TRANSMIT_ENA_OUT <= 1'b0;
				transmit_fifo_wren <= 1'b0;
				transmit_fifo_wren_reg <= 1'b0;
			end
		end
end


//=================================================================================
//	Write Cycle: Send word from transmit FIFO to transmit unit
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			transmit_data_valid <= 1'b0;
			transmit_data_valid_reg <= 2'b00;
			transmit_data_complete <= 1'b0;
			transmit_fifo_rden <= 1'b0;
			transmit_active <= 1'b0;
			transmit_data_complete <= 1'b0;
		end
		else
		begin
			//if(state[WRITE_BYTE] & !transmit_fifo_empty & (transmit_data_valid_reg == 2'b00))
			if(state[WRITE_BYTE] & (transmit_data_valid_reg == 2'b00))
			begin
					transmit_data_valid <= 1'b0;
					transmit_fifo_rden <= 1'b1;
					transmit_data_valid_reg <= 2'b01;
					transmit_active <= 1'b1;
					transmit_data_complete <= 1'b0;
			end	
			else if(state[WRITE_BYTE] & (transmit_data_valid_reg == 2'b01))
			begin
					transmit_data_valid <= 1'b1;
					transmit_fifo_rden <= 1'b1;
					transmit_data_valid_reg <= 2'b10;
			end
			else if(state[WRITE_BYTE] & (transmit_data_valid_reg == 2'b10))
			begin
					transmit_data_valid <= 1'b1;
					transmit_fifo_rden <= 1'b0;
					transmit_data_valid_reg <= 2'b11;
					transmit_active <= 1'b0;
			end
			else if(state[WRITE_ACK] & (transmit_data_valid_reg == 2'b11))
			begin
					transmit_data_valid <= 1'b0;
					transmit_fifo_rden <= 1'b0;
					transmit_data_valid_reg <= 2'b11;
					transmit_data_complete <= 1'b1;
					transmit_active <= 1'b0;
				if ( !transmit_active )
					transmit_data_complete <= 1'b1;
			end
			else if( state[IDLE] )
			begin
				transmit_data_valid <= 1'b0;
				transmit_fifo_rden <= 1'b0;
				transmit_data_valid_reg <= 1'b0;
				transmit_data_complete <= 1'b0;
				transmit_data_complete <= 1'b0;
				transmit_active <= 1'b0;
			end
		end
end

//=================================================================================
//	Receive Cycle: Set the receive_active signal to inducate to the I2C receive
// unit to start decoding serial input
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if( !RESET )
		begin
			receive_active <= 1'b0;
			receive_active_reg <= 2'b00;
		end
		else
		begin
			if((state[I2C_START] | state[READ_BYTE]) & (receive_active_reg == 2'b00))
			begin
				receive_active <= 1'b1;
				receive_active_reg <= 2'b01;
			end
			if((state[I2C_START] | state[READ_BYTE]) & (receive_active_reg == 2'b01))
			begin
				receive_active <= 1'b1;
				receive_active_reg <= 2'b10;
			end
			if((state[I2C_START] | state[READ_BYTE]) & (receive_active_reg == 2'b10))
			begin
				receive_active <= 1'b0;
				receive_active_reg <= 2'b11;
			end
			else if (state[READ_ADDRESS] | state[WRITE_BYTE] | state[READ_ACK_DELAY] | state[I2C_STOP] )
			begin
				receive_active <= 1'b0;
				receive_active_reg <= 2'b00;
			end
		end
	end
	

//=================================================================================
//	Receive Cycle: Store the I2C address from the Master
//=================================================================================

	always @( negedge CLK or negedge RESET )
	begin
		if( !RESET )
		begin
			i2c_address_detected <= 1'b0;
		end
		else
		begin
			if( i2c_address[7:1] == DEVICE_ADDRESS[7:1] )
				i2c_address_detected <= 1'b1;
			else if ( state[IDLE] )
			begin
					i2c_address_detected <= 1'b0;
			end
		end
	end


 //=================================================================================
 //	Receive Cycle: Detect Read Write Bit
 //=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if( !RESET )
		begin
			i2c_read_write_bit <= 1'b0;
		end
		else
		begin
			if(i2c_address[0])
				i2c_read_write_bit <= 1'b1;
			else if ( state[IDLE] )
			begin
					i2c_read_write_bit <= 1'b0;
			end
		end
	end


//=================================================================================
//	Receive Cycle: Assert the SERIAL_IIC_RECEIVE_RDY_OUT register when receive FIFO has data
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			SERIAL_IIC_RECEIVE_RDY_OUT <= 1'b0;
			serial_iic_rdy_out_reg <= 2'b00;
		end
		else
		begin
			if(!receive_fifo_empty & (serial_iic_rdy_out_reg == 2'b00))
			begin
				SERIAL_IIC_RECEIVE_RDY_OUT <= 1'b1;
				serial_iic_rdy_out_reg <= 2'b01;
			end
			else if(SERIAL_IIC_RECEIVE_ENA_IN & (serial_iic_rdy_out_reg == 2'b01))
			begin
				SERIAL_IIC_RECEIVE_RDY_OUT <=1'b0;
				serial_iic_rdy_out_reg <= 2'b10;
			end
			else if (!SERIAL_IIC_RECEIVE_ENA_IN & (serial_iic_rdy_out_reg == 2'b10))
			begin
				SERIAL_IIC_RECEIVE_RDY_OUT <=1'b0;
				serial_iic_rdy_out_reg <= 2'b00;
			end
		end
	end


//=================================================================================
//	Receive Cycle: Read word from receive FIFO when fifo_rden is asserted
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			receive_fifo_rden <= 1'b0;
			receive_fifo_rden_reg <= 1'b0;
		end
		else
		begin
			if( SERIAL_IIC_RECEIVE_ENA_IN & !receive_fifo_rden_reg )
			begin
				receive_fifo_rden <= 1'b1;
				receive_fifo_rden_reg <= 1'b1;
			end
			else if( SERIAL_IIC_RECEIVE_ENA_IN & receive_fifo_rden_reg )
			begin
				receive_fifo_rden <= 1'b0;
				receive_fifo_rden_reg <= 1'b1;
			end
			else if( !SERIAL_IIC_RECEIVE_ENA_IN & receive_fifo_rden_reg  )
			begin
				receive_fifo_rden <= 1'b0;
				receive_fifo_rden_reg <= 1'b0;
			end
		end
	end


//=================================================================================
//	Receive Cycle: Load valid word from receive unit into FIFO when valid_out is asserted
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			receive_fifo_wren <= 1'b0;
			receive_fifo_wren_reg <= 1'b0;
			wait_for_data_valid  <= 1'b0;
		end
		else
		begin
			if( receive_valid_out & state[READ_BYTE] & !wait_for_data_valid & !receive_fifo_wren_reg )
			begin
				wait_for_data_valid <= 1'b1;
				receive_fifo_wren_reg <= 1'b0;
			end
			else if( !receive_valid_out & state[READ_BYTE] & wait_for_data_valid & !receive_fifo_wren_reg )
			begin
				receive_fifo_wren <= 1'b0;
				receive_fifo_wren_reg <= 1'b1;
				wait_for_data_valid <= 1'b1;
			end
			else if( receive_valid_out & state[READ_BYTE] & wait_for_data_valid & receive_fifo_wren_reg )
			begin
				receive_fifo_wren <= 1'b1;
				receive_fifo_wren_reg <= 1'b1;
				wait_for_data_valid <= 1'b1;
			end
			else if( state[READ_ACK])
			begin
				receive_fifo_wren <= 1'b0;
				receive_fifo_wren_reg <= 1'b0;
				wait_for_data_valid  <= 1'b1;
			end
			else if( state[IDLE])
			begin
				receive_fifo_wren <= 1'b0;
				receive_fifo_wren_reg <= 1'b0;
				wait_for_data_valid  <= 1'b0;
			end
		end
	end
		
//=================================================================================
//	Receive Cycle: Create the receive_decode_message for the receive unit 
//=================================================================================
/*	always @( posedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			receive_decode_message <= 1'b0;
			receive_decode_message_reg  <= 1'b0;
		end
		else
		begin
			if( receive_active & !receive_decode_message_reg )
			begin
				receive_decode_message <= 1'b1;
				receive_decode_message_reg  <= 1'b1;
			end
			else if( receive_active & receive_decode_message_reg )
			begin
				receive_decode_message <= 1'b0;
				receive_decode_message_reg  <= 1'b1;
			end
			else if( !receive_active & receive_decode_message_reg )
			begin
				receive_decode_message <= 1'b0;
				receive_decode_message_reg  <= 1'b0;
			end
			else if( state[IDLE] )
				receive_decode_message <= 1'b0;

		end
	end
*/
//=================================================================================
//	Ack Send Cycle: Ack Grant Register
//=================================================================================


	always @( negedge CLK or negedge RESET )
	begin
		if ( !RESET )
		begin
			ack_grant_start <= 1'b0;			
			ack_grant_start_reg <= 1'b0;
		end
		else
		begin
			if(receive_active & !ack_grant_start_reg)
			begin
				ack_grant_start <= 1'b1;
				ack_grant_start_reg <= 1'b1;
			end
			else if(ack_done & ack_grant_start_reg)
			begin
				ack_grant_start <= 1'b0;
				ack_grant_start_reg <= 1'b1;
			end
			else if(!receive_active & ack_grant_start_reg)
			begin
				ack_grant_start <= 1'b0;
				ack_grant_start_reg <= 1'b0;
			end
		end
end

	
//=================================================================================
//	Serial I2C Communications Finite State Machine
//
//
//
//=================================================================================

	always @(  posedge CLK or negedge RESET  )
	begin
		if( !RESET )
		begin
			state <= 12'h00;
			state [IDLE] <= 1'b1;
		end
		else
			state <= next;
	end


	always @( state or i2c_start_detect or receive_valid_out or 
		    ack_done or i2c_address_detected or i2c_read_write_bit or 
		    i2c_rw_stop /*or transmit_holding_empty */or transmit_done or
			rstop_phase or i2c_next_byte_timeout_cntr or i2c_next_byte_flag or
			I2C_CLOCK)
	begin
		next = 12'h00;
		if (state[IDLE])
		begin
			if (i2c_start_detect)
				next[I2C_START] = 1'b1;
			else
				next[IDLE] = 1'b1;
		end

		if (state[I2C_START])
		begin
			if (receive_valid_out)
				next[READ_ADDRESS] = 1'b1;
			else
				next[I2C_START] = 1'b1;
		end

		if (state[READ_ADDRESS])
		begin
			if (i2c_address_detected)
				next[DECODE_RW] = 1'b1;
			else
				next[I2C_STOP] = 1'b1;
		end

		if (state[DECODE_RW])
		begin
		    if(!ack_done)
				next[READ_ACK] = 1'b1;
			else
				next[DECODE_RW] = 1'b1;
		end

		if (state[CHECK_READ_WRITE])
		begin
		    if(i2c_rw_stop)
			    next[I2C_STOP] = 1'b1;
		    else if(i2c_next_byte_flag)
			begin
		       if (i2c_read_write_bit)
		          next[WRITE_BYTE] = 1'b1;
		       else
		          next[READ_BYTE] = 1'b1;
			end
			else
				next[CHECK_READ_WRITE] = 1'b1;
		end

		if (state[WRITE_BYTE])
		begin
		    if(i2c_rw_stop)
				next[I2C_STOP] = 1'b1;
			else if (i2c_ack_rcv_active)
				next[WRITE_ACK] = 1'b1;
			else
				next[WRITE_BYTE] = 1'b1;
		end

		if (state[WRITE_ACK])
		begin
			if (ack_done)
				next[WRITE_ACK_DELAY] = 1'b1;
			else
				next[WRITE_ACK] = 1'b1;
		end

		if (state[WRITE_ACK_DELAY])
		begin
			if (!ack_done)
				next[WRITE_BYTE] = 1'b1;
			else
				next[WRITE_ACK_DELAY] = 1'b1;
		end

		if (state[READ_BYTE])
		begin
		    if(i2c_rw_stop)
				next[I2C_STOP] = 1'b1;
			else if (receive_valid_out)
				next[READ_ACK] = 1'b1;
			else
				next[READ_BYTE] = 1'b1;
		end

		if (state[READ_ACK])
		begin
			if (ack_done)
				next[READ_ACK_DELAY] = 1'b1;
			else
				next[READ_ACK] = 1'b1;
		end

		if (state[READ_ACK_DELAY])
		begin
			if (!ack_done & !I2C_CLOCK)
				//next[READ_BYTE] = 1'b1;
				next[CHECK_READ_WRITE] = 1'b1;
			else
				next[READ_ACK_DELAY] = 1'b1;
		end

		if (state[I2C_STOP])
		begin
			//if (i2c_rw_stop)
				next[IDLE] = 1'b1;
			//else
			//	next[I2C_STOP] = 1'b1;
		end


//=================================================================
// For Simumlation, display the state names
//
//	parameter	IDLE                         = 0,
//	            I2C_START                    = 1,
//              READ_ADDRESS                 = 2,
//				DECODE_RW                    = 3,
//              CHECK_READ_WRITE             = 4,
//				WRITE_BYTE                   = 5,
//				WRITE_ACK                    = 6,
//              WRITE_ACK_DELAY              = 7,
//				READ_BYTE                    = 8,
//				READ_ACK                     = 9,
//              READ_ACK_DELAY               = 10,
//				I2C_STOP                     = 11;
//
//=================================================================

`ifdef SIM
	if ( state == (1 << IDLE))
		state_name = "IDLE";		
	if ( state == (1 << I2C_START))
		state_name = "I2C_START";		
	if ( state == (1 << READ_ADDRESS))
		state_name = "READ_ADDRESS"; 		
	if ( state == (1 << DECODE_RW))
		state_name = "DECODE_RW";		
	if ( state == (1 << CHECK_READ_WRITE))
		state_name = "CHECK_READ_WRITE";		
	if ( state == (1 << WRITE_BYTE))
		state_name = "WRITE_BYTE";		
	if ( state == (1 << WRITE_ACK))
		state_name = "WRITE_ACK";
	if ( state == (1 << WRITE_ACK_DELAY))
		state_name = "WRITE_ACK_DELAY";
	if ( state == (1 << READ_BYTE))
		state_name = "READ_BYTE";
	if ( state == (1 << READ_ACK))
		state_name = "READ_ACK";
	if ( state == (1 << READ_ACK_DELAY))
		state_name = "READ_ACK_DELAY";
	if ( state == (1 << I2C_STOP))
		state_name = "I2C_STOP";
//end
`endif		
	end


//==============================================================================
//	Instantiate the IIC Transmit Interface
//==============================================================================
    i2c_transmit                 SERIAL_I2C_TRANSMIT_INST
	(
		.clk					(CLK),
		.reset                  (RESET),

		.iic_data				(iic_data_from_transmit),
		.iic_clock				(I2C_CLOCK),

		.transmit_complete		(transmit_done),
		.data_in				(transmit_data_in),
		.data_valid				(transmit_data_valid),
		.active                 (transmit_active)
	);


//==============================================================================
//	Instantiate the IIC Receive Interface
//==============================================================================
    i2c_receive                 SERIAL_I2C_RECEIVE_INST
	(
		.CLK					(CLK),
		.RST_N                  (RESET),

		.DATA_IN				(IIC_DATA),
		.SCLK				    (I2C_CLOCK),

		.DOUT				    (receive_data_out),
		.DONE				    (receive_valid_out),
		.START                  (receive_active),
		
		.PERFORM_READ           (),
		
		.TEST_BUS               (i2c_receive_out_phase)
	);

  //==============================================================================
  //	Instantiate the ACK Unit
  //==============================================================================
    i2c_ack                     SERIAL_I2C_ACK_INST
	(
		.CLK					(CLK),
		.RST_N                  (RESET),

		.GNT_START				(ack_grant_start),
		.REQ_START				(ack_request_start),
		.RECEIVE_UNIT_DONE      (receive_valid_out),
		.TRANSMIT_UNIT_DONE     (transmit_done),
        .VALID_SEND             (i2c_ack_send_active),
	    .VALID_RECEIVE          (i2c_ack_rcv_active),

		.DATA_IN				(i2c_ack_transmit),
		.DOUT					(ack_data_out),
		.DATA_OUT				(i2c_data_out_from_ack),
		.DONE					(ack_done),
		.SCLK					(i2c_ack_clock),
		
		.I2C_STOP               (i2c_rw_stop),

        .TEST_BUS_OUT                (i2c_ack_test_bus_out),
		.STATE_OUT                   (i2c_ack_state)
	);


//-----------------------------------------------
// Instantiate FIFO for receiving incoming data 
//	from IIC bus fifo2048x8  module
//-----------------------------------------------
   
  sync_fifo #(
      .ADDR_WIDTH               (3),
	  .DATA_WIDTH               (8),
	  .DEPTH                    (8),
	  .AEMPTY                   (1),
	  .AFULL                    (6)
	  ) I2C_RECEIVE_FIFO
     (
	  .clk                     (CLK),
	  .reset_n                 (RESET),
	  .flush                   (receive_fifo_flush),
	  .write_data              (receive_fifo_din),	  
	  .read_req                (receive_fifo_rden),
	  .wdata_valid             (receive_fifo_wren),
	  .read_data               (receive_fifo_dout),
	  .rdata_valid             (),
	  .fifo_empty              (receive_fifo_empty),
	  .fifo_aempty             (receive_fifo_aempty),
	  .fifo_full               (receive_fifo_full),
	  .fifo_afull              (receive_fifo_afull),
	  .write_ack               (receive_fifo_wr_ack)
      );

 
//-----------------------------------------------
// Instantiate FIFO for storing outgoing data 
//	to the IIC bus fifo2048x8  module
//-----------------------------------------------
 
  sync_fifo #(
      .ADDR_WIDTH               (3),
	  .DATA_WIDTH               (8),
	  .DEPTH                    (8),
	  .AEMPTY                   (1),
	  .AFULL                    (6)
	  ) I2C_TRANSMIT_FIFO
     (
	  .clk                     (CLK),
	  .reset_n                 (RESET),
	  .flush                   (transmit_fifo_flush),
	  .write_data              (transmit_fifo_din),	  
	  .read_req                (transmit_fifo_rden),
	  .wdata_valid             (transmit_fifo_wren),
	  .read_data               (transmit_fifo_dout),
	  .rdata_valid             (),
	  .fifo_empty              (transmit_fifo_empty),
	  .fifo_aempty             (transmit_fifo_aempty),
	  .fifo_full               (transmit_fifo_full),
	  .fifo_afull              (transmit_fifo_afull),
	  .write_ack               (transmit_fifo_wr_ack)
      );

endmodule

